#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class RecvBuffSizeComputer:
    def __init__(self):
        self.min_recv_buff_size = 1024
        self.max_recv_buff_size = 1024 * 1024 * 10
        self.recv_buff_size = self.min_recv_buff_size
        self._last_recv_amount = 0
        self._previous_recv_amount = 0

    def calc_new_recv_buff_size(self, last_recv_amount):
        biggest_recv_amount = 0
        new_amount = 0

        # moving resv amounts on list
        self._previous_recv_amount = self._last_recv_amount
        self._last_recv_amount = last_recv_amount

        new_amount = self.recv_buff_size
        if (self._last_recv_amount >= self.recv_buff_size) \
                and (self._previous_recv_amount >= self.recv_buff_size):
            new_amount = self.recv_buff_size << 2
        elif (self._last_recv_amount >= self.recv_buff_size) \
                and (self._previous_recv_amount < self.recv_buff_size):
            new_amount = self.recv_buff_size << 1
        elif (self._last_recv_amount < self.recv_buff_size) \
                and (self._previous_recv_amount >= self.recv_buff_size):
            new_amount = self.recv_buff_size
        elif (self._last_recv_amount < self.recv_buff_size) \
                and (self._previous_recv_amount < self.recv_buff_size):
            # new_amount = max(self._last_recv_amounts_list) * 1.5
            if self._last_recv_amount > self._previous_recv_amount:
                biggest_recv_amount = self._last_recv_amount
            else:
                biggest_recv_amount = self._previous_recv_amount
            new_amount = biggest_recv_amount + (biggest_recv_amount >> 1)

        if new_amount < self.min_recv_buff_size:
            self.recv_buff_size = self.min_recv_buff_size
        elif new_amount > self.max_recv_buff_size:
            self.recv_buff_size = self.max_recv_buff_size
        else:
            self.recv_buff_size = new_amount

        return self.recv_buff_size



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['SubprocessIsNotInitiatedError', 'SubprocessIsNotReadyError', 'SubprocessTerminatedError', 'Empty', 'Full', 'SendableDataType', 'Transport', 'SubprocessWorkerSettings', 'SubprocessWorker', '_subprocess_wrapper_profile', 'ExternalPipe']


import cProfile
# import multiprocessing
from multiprocessing import Process, Queue, Pipe
from threading import Thread
import sys
import traceback
from queue import Empty, Full
import marshal
import pickle
import os
from cengal.data_manipulation.front_triggerable_variable import FrontTriggerableVariable, FrontTriggerableVariableType
from cengal.base.classes import BaseClassSettings
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.introspection.inspect import pdi
from typing import Callable, Any, Union, Optional, Tuple, List, Dict

# import time

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class SubprocessIsNotInitiatedError(Exception):
    pass


class SubprocessIsNotReadyError(Exception):
    pass


class SubprocessTerminatedError(Exception):
    pass


class SendableDataType:
    pickable = 0
    marshalable = 1
    custom = 2


class Transport:
    queue = 0
    pipe = 1
    tcp = 2


class SubprocessWorkerSettings(BaseClassSettings):
    def __init__(self):
        self.initiation_function = None  # self.data_shelf = self.settings.initiation_function(
        #   self.initialization_data)
        self.working_function = None  # 1) 'answer = self.settings.working_function(input_data)'
        #       if initiation_function is None;
        #   2) 'answer = self.settings.working_function(self.data_shelf, input_data)' if initiation_function
        #       is not None.
        self.stopping_function = None

        self.on_input_queue_is_too_big = None  # self.on_input_queue_is_too_big(self.data_shelf,
        #   average_input_size_trigger_result) where: average_input_size_trigger_result =
        #   self.input_queue_average_size_trigger.test_trigger(average_input_size)
        self.on_another_bunch_of_data_was_processed = None  # self.on_another_bunch_of_data_was_processed(
        #   self.data_shelf)
        self.on_exit = None  # on process exit; self.on_exit(self.data_shelf)

        self.need_multithreading = False  # will use multithread mode if True; and multiprocess else.
        self.process_name: str = None  # str()
        self.profile = False  # will start worker in profiling mode
        self.initialization_data = None  # any pickable Python data
        self.transport: Transport = None  # Transport()
        self.tcp_settings = None  # tcp_link.TCPSettings(). Is used when (self.transport == Transport.tcp)
        self.use_internal_subprocess_input_buffer = False  # will be able to get input data in nonblocking mode
        self.sendable_data_type: SendableDataType = SendableDataType.pickable
        self.sendable_data__encoder = None  # Will be in use when sendable_data_type == SendableDataType.custom.
        #   Function will encode data in to bytes()
        self.sendable_data__decoder = None  # Will be in use when sendable_data_type == SendableDataType.custom.
        #   Function will decode data from bytes()
        self.queue_to_subprocess = None  # only needed if you want to directly connect this worker with another.
        #   For example to connect output from this worker to input of another worker.
        self.queue_from_subprocess = None  # only needed if you want to directly connect this worker with another.
        #   For example to connect output from this worker to input of another worker.
        self.input_queue_average_size_trigger_limit = 30
        self.subprocess_polling_timeout = 0.0  # None - infinite; 0.0 - nonblocking; > 0.0 - timeout in seconds
        self.subprocess_reading_timeout = 0.1  # None - infinite; 0.0 - nonblocking; > 0.0 - timeout in seconds
        self.subprocess_writing_timeout = 0.1  # None - infinite; 0.0 - nonblocking; > 0.0 - timeout in seconds
        self.subprocess_invalidation_timeout = 0.1  # None - infinite; 0.0 - nonblocking; > 0.0 - timeout in seconds
        self.indicate_subprocess_readyness: bool = True

    def check(self):
        if self.working_function is None:
            raise Exception('working_function cannot be None!')

        if self.transport is None:
            raise Exception('transport can\'t be None')
        else:
            if Transport.tcp == self.transport:
                if self.tcp_settings is None:
                    raise Exception('tcp_settings can\'t be None while (Transport.tcp == self.transport)')
                else:
                    self.tcp_settings.check()


class SubprocessWorker:
    def __init__(self, settings: SubprocessWorkerSettings):
        """ 
        :param settings: SubprocessWorkerSettings(); you should use copy.copy(SubprocessWorkerSettings(...)) by your 
            self if you want
        :return: 
        """
        super().__init__()
        
        self.settings: SubprocessWorkerSettings = settings
        self.settings.check()

        self.data_shelf = None

        self.subprocess_was_initiated = False
        self.subprocess = None
        self.queue_to_subprocess = self.settings.queue_to_subprocess
        self.queue_from_subprocess = self.settings.queue_from_subprocess

        self.list_of_subprocess_input_data: List = list()
        self.input_size_print_sum = 0
        self.input_size_print_counter = 0
        self.input_size_print_counter_limit = 500
        # self.last_log_print_time = time.time()
        self.subprocess_readyness_indicated: bool = False
        self.input_queue_average_size_trigger = FrontTriggerableVariable(
            FrontTriggerableVariableType.bigger_or_equal, self.settings.input_queue_average_size_trigger_limit)

    def _encode_sendable_data(self, data):
        result = data
        if Transport.queue == self.settings.transport:
            if self.settings.sendable_data_type == SendableDataType.pickable:
                pass
            elif self.settings.sendable_data_type == SendableDataType.marshalable:
                result = marshal.dumps(result)
            elif self.settings.sendable_data_type == SendableDataType.custom:
                result = self.settings.sendable_data__encoder(result)
        else:
            if self.settings.sendable_data_type == SendableDataType.pickable:
                result = pickle.dumps(result)
            elif self.settings.sendable_data_type == SendableDataType.marshalable:
                result = marshal.dumps(result)
            elif self.settings.sendable_data_type == SendableDataType.custom:
                result = self.settings.sendable_data__encoder(result)
        return result

    def _decode_sendable_data(self, data):
        result = data
        if Transport.queue == self.settings.transport:
            if self.settings.sendable_data_type == SendableDataType.pickable:
                pass
            elif self.settings.sendable_data_type == SendableDataType.marshalable:
                result = marshal.loads(result)
            elif self.settings.sendable_data_type == SendableDataType.custom:
                result = self.settings.sendable_data__decoder(result)
        else:
            if self.settings.sendable_data_type == SendableDataType.pickable:
                result = pickle.loads(result)
            elif self.settings.sendable_data_type == SendableDataType.marshalable:
                result = marshal.loads(result)
            elif self.settings.sendable_data_type == SendableDataType.custom:
                result = self.settings.sendable_data__decoder(result)
        return result

    def _parsing_worker_wrapper(self, input_data, stop=False):
        # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '0')
        exception = None
        answer = None
        result = None
        try:
            # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '1')
            if self.settings.initiation_function is None:
                # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '2')
                if stop:
                    if self.settings.stopping_function is not None:
                        answer = self.settings.stopping_function()
                    else:
                        answer = 'Stopped'
                else:
                    answer = self.settings.working_function(input_data)
                # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '3')
            else:
                # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '4')
                if stop:
                    if self.settings.stopping_function is not None:
                        answer = self.settings.stopping_function(self.data_shelf)
                    else:
                        answer = 'Stopped'
                else:
                    answer = self.settings.working_function(self.data_shelf, input_data)
                # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '5')
        except:
            # # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '6')
            # exception = sys.exc_info()
            # formatted_traceback = traceback.format_exception(exception[0], exception[1], exception[2])
            # exception = exception[:2] + (formatted_traceback,)
            # answer = (input_data[0], None)
            # # print(self.settings.process_name)
            # # print(input_data)
            # # print(exception)
            answer = None
            exception = pickle.dumps(sys.exc_info())

        # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '7')
        if (answer is not None) or (exception is not None):
            # print('===>>', self.settings.process_name, '_parsing_worker_wrapper', '8')
            result = (answer, exception)

        # print('<<===', self.settings.process_name, '_parsing_worker_wrapper')
        return result

    def _subprocess_wrapper(self):
        try:
            # input_from_parent_process_queue = self.queue_to_subprocess
            # output_to_parent_process_queue = self.queue_from_subprocess

            # print(' STARTED:', self.settings.process_name, '; PID:', os.getpid())

            input_from_parent_process_queue = None
            output_to_parent_process_queue = None
            if Transport.pipe == self.settings.transport:
                input_from_parent_process_queue = self.queue_to_subprocess[1]
                output_to_parent_process_queue = self.queue_from_subprocess[1]
            elif Transport.queue == self.settings.transport:
                input_from_parent_process_queue = self.queue_to_subprocess
                output_to_parent_process_queue = self.queue_from_subprocess

            if self.settings.indicate_subprocess_readyness:
                result = (True, ('Started', None))
                result = self._encode_sendable_data(result)
                if Transport.pipe == self.settings.transport:
                    output_to_parent_process_queue.send_bytes(result)
                elif Transport.queue == self.settings.transport:
                    output_to_parent_process_queue.put(result)

            if self.settings.initiation_function is not None:
                self.data_shelf = self.settings.initiation_function(self.settings.initialization_data)
            while True:
                # print('===>>', self.settings.process_name, '_subprocess_wrapper', '0')
                # print('===>>', self.settings.process_name, '_subprocess_wrapper', '1', 'qsize:',
                #       input_from_parent_process_queue.qsize())

                # input_size = input_from_parent_process_queue.qsize()
                # if input_size > 3:
                #     print('===>>', self.settings.process_name, 'input_size:', input_size)
                # output_size = output_to_parent_process_queue.qsize()
                # if output_size > 3:
                #     print('===>>', self.settings.process_name, 'output_size:', output_size)

                data = None
                if self.settings.use_internal_subprocess_input_buffer:
                    while True:
                        another_chunk_of_data = None
                        if Transport.pipe == self.settings.transport:
                            if input_from_parent_process_queue.poll():
                                another_chunk_of_data = input_from_parent_process_queue.recv_bytes()
                            else:
                                break
                        elif Transport.queue == self.settings.transport:
                            if not input_from_parent_process_queue.empty():
                                another_chunk_of_data = input_from_parent_process_queue.get()
                            else:
                                break
                        
                        if another_chunk_of_data:
                            self.list_of_subprocess_input_data.append(another_chunk_of_data)

                    input_size = len(self.list_of_subprocess_input_data)
                    self.input_size_print_counter += 1
                    self.input_size_print_sum += input_size
                    if self.input_size_print_counter >= self.input_size_print_counter_limit:
                        average_input_size = self.input_size_print_sum / self.input_size_print_counter
                        self.input_size_print_sum = 0
                        self.input_size_print_counter = 0
                        average_input_size_trigger_result = self.input_queue_average_size_trigger.test_trigger(
                            average_input_size)
                        if average_input_size_trigger_result is not None:
                            # if average_input_size_trigger_result:
                            #     print('===>>', self.settings.process_name, 'average_input_size:', average_input_size)
                            # else:
                            #     print('===>>', self.settings.process_name, 'average_input_size is OK:', average_input_size)

                            if self.settings.on_input_queue_is_too_big is not None:
                                self.settings.on_input_queue_is_too_big(self.data_shelf, average_input_size_trigger_result)

                    # data = input_from_parent_process_queue.get(block=False, timeout=self.settings.subprocess_reading_timeout)
                    if len(self.list_of_subprocess_input_data) > 0:
                        data = self.list_of_subprocess_input_data[0]
                        # print('===>>', self.settings.process_name, 'input_data:', data)
                        self.list_of_subprocess_input_data = self.list_of_subprocess_input_data[1:]
                    else:
                        continue
                else:
                    if Transport.pipe == self.settings.transport:
                        data = input_from_parent_process_queue.recv_bytes()
                    elif Transport.queue == self.settings.transport:
                        input_size = input_from_parent_process_queue.qsize()
                        self.input_size_print_counter += 1
                        self.input_size_print_sum += input_size
                        if self.input_size_print_counter >= self.input_size_print_counter_limit:
                            average_input_size = self.input_size_print_sum / self.input_size_print_counter
                            self.input_size_print_sum = 0
                            self.input_size_print_counter = 0
                            average_input_size_trigger_result = self.input_queue_average_size_trigger.test_trigger(
                                average_input_size)
                            if average_input_size_trigger_result is not None:
                                # if average_input_size_trigger_result:
                                #     print('===>>', self.settings.process_name, 'average_input_size for Queue:',
                                #         average_input_size)
                                # else:
                                #     print('===>>', self.settings.process_name, 'average_input_size for Queue is OK:',
                                #         average_input_size)
                                
                                if self.settings.on_input_queue_is_too_big is not None:
                                    self.settings.on_input_queue_is_too_big(self.data_shelf,
                                                                            average_input_size_trigger_result)

                        try:
                            data = input_from_parent_process_queue.get(block=True)
                        except Empty:
                            pass

                # print('===>>', self.settings.process_name, '_subprocess_wrapper', '2')

                # current_time = time.time()
                # if (current_time - self.last_log_print_time) > 2:
                #     print('===>>', self.settings.process_name, '_subprocess_wrapper')

                if data is None:
                    continue

                is_result_was_send = False
                is_worker_is_finalized = False
                is_need_to_break_loop = False

                data = self._decode_sendable_data(data)
                # data = marshal.loads(data)
                continue_processing: bool = data[0]
                if continue_processing:
                    # print('===>>', self.settings.process_name, '_subprocess_wrapper', '3')
                    data_with_exception = data[1]
                    data_only = data_with_exception[0]
                    result = self._parsing_worker_wrapper(data_only)
                    if result is not None:
                        # print('===>>', self.settings.process_name, 'output_result:', result)
                        # print('===>>', self.settings.process_name, '_subprocess_wrapper', '5')
                        result = (True, result)
                        result = self._encode_sendable_data(result)
                        # result = marshal.dumps(result)
                        # output_to_parent_process_queue.put(result)
                        if Transport.pipe == self.settings.transport:
                            output_to_parent_process_queue.send_bytes(result)
                        elif Transport.queue == self.settings.transport:
                            output_to_parent_process_queue.put(result)

                        is_result_was_send = True
                    # print('===>>', self.settings.process_name, '_subprocess_wrapper', '5')
                else:
                    # print('===>>', self.settings.process_name, '_subprocess_wrapper', '6')

                    data_only = (None, None)
                    self._parsing_worker_wrapper(data_only, stop=True)
                    if result is not None:
                        # print('===>>', self.settings.process_name, 'output_result:', result)
                        # print('===>>', self.settings.process_name, '_subprocess_wrapper', '5')
                        result = (False, result)
                        result = self._encode_sendable_data(result)
                        # result = marshal.dumps(result)
                        # output_to_parent_process_queue.put(result)
                        if Transport.pipe == self.settings.transport:
                            output_to_parent_process_queue.send_bytes(result)
                        elif Transport.queue == self.settings.transport:
                            output_to_parent_process_queue.put(result)

                        is_result_was_send = True

                    is_worker_is_finalized = True
                    is_need_to_break_loop = True
                    # break

                if continue_processing and (self.settings.on_another_bunch_of_data_was_processed is not None) and is_result_was_send:
                    self.settings.on_another_bunch_of_data_was_processed(self.data_shelf)

                if (self.settings.on_exit is not None) and is_worker_is_finalized:
                    self.settings.on_exit(self.data_shelf)

                if is_need_to_break_loop:
                    break

                # print('<<===', self.settings.process_name, '_subprocess_wrapper')

                # if (current_time - self.last_log_print_time) > 2:
                #     print('<<===', self.settings.process_name, '_subprocess_wrapper')
                #     self.last_log_print_time = current_time

            # print(' ENDED:', self.settings.process_name, '; PID:', os.getpid())
            # print('<<===>>', self.settings.process_name, '_subprocess_wrapper')
        except BrokenPipeError:
            pass
        except OSError:
            pass
        except ValueError:
            pass
        finally:
            if Transport.pipe == self.settings.transport:
                input_from_parent_process_queue = self.queue_to_subprocess[1]
                output_to_parent_process_queue = self.queue_from_subprocess[1]
            elif Transport.queue == self.settings.transport:
                input_from_parent_process_queue = self.queue_to_subprocess
                output_to_parent_process_queue = self.queue_from_subprocess
            
            input_from_parent_process_queue.close()
            output_to_parent_process_queue.close()
            

    def start(self, wait_for_process_readyness: bool = True):
        if not self.subprocess_was_initiated:
            if self.queue_to_subprocess is None:
                if Transport.pipe == self.settings.transport:
                    self.queue_to_subprocess = Pipe()
                elif Transport.queue == self.settings.transport:
                    self.queue_to_subprocess = Queue()

            if self.queue_from_subprocess is None:
                if Transport.pipe == self.settings.transport:
                    self.queue_from_subprocess = Pipe()
                elif Transport.queue == self.settings.transport:
                    self.queue_from_subprocess = Queue()

            target = None
            arguments = None
            if self.settings.profile:
                target = _subprocess_wrapper_profile
                arguments = (self,)
            else:
                target = self._subprocess_wrapper
                arguments = tuple()
            self.subprocess = None
            if self.settings.need_multithreading:
                self.subprocess = Thread(target=target, args=arguments, daemon=True)
            else:
                self.subprocess = Process(target=target, args=arguments, daemon=True)
            
            self.subprocess.start()
            self.subprocess_was_initiated = True
        
        if wait_for_process_readyness:
            self.wait_for_subprocess_readines(block=True)
    
    def wait_for_subprocess_readines(self, block: bool = True):
        if not self.subprocess_was_initiated:
            raise SubprocessIsNotInitiatedError
        
        if self.subprocess_readyness_indicated:
            return

        if self.settings.indicate_subprocess_readyness:
            self.subprocess_readyness_indicated = True
            try:
                try:
                    self.get_answer_from_subprocess(block=block)
                    self.subprocess_readyness_indicated = True
                except Empty:
                    raise SubprocessIsNotReadyError
            except:
                self.subprocess_readyness_indicated = False
                raise
    
    def _close_connections(self):
        if self.subprocess_was_initiated:
            if Transport.pipe == self.settings.transport:
                self.queue_to_subprocess[0].close()
                self.queue_from_subprocess[0].close()
            elif Transport.queue == self.settings.transport:
                self.queue_to_subprocess.close()
                self.queue_from_subprocess.close()

    def stop(self):
        if not self.subprocess_was_initiated:
            return
        
        self.wait_for_subprocess_readines(block=True)

        data = (False, (None, None))
        data = self._encode_sendable_data(data)
        # data = marshal.dumps(data)

        output_to_subprocess_queue = None
        if Transport.pipe == self.settings.transport:
            output_to_subprocess_queue = self.queue_to_subprocess[0]
        elif Transport.queue == self.settings.transport:
            output_to_subprocess_queue = self.queue_to_subprocess

        if Transport.pipe == self.settings.transport:
            output_to_subprocess_queue.send_bytes(data)
        elif Transport.queue == self.settings.transport:
            output_to_subprocess_queue.put(data, timeout=self.settings.subprocess_writing_timeout)
        
        try:
            self.get_answer_from_subprocess(block=True)
        except Empty:
            pass
        except SubprocessTerminatedError:
            pass
        finally:
            self._close_connections()
            self.subprocess_was_initiated = False
            self.subprocess.join()
    
    def _invalidate(self):
        self._close_connections()
        self.subprocess_was_initiated = False
        self.subprocess.join(self.settings.subprocess_invalidation_timeout)

    def send_data_to_subprocess(self, input_data, block: bool = True):
        """
        If (Transport.pipe == self.settings.transport): Very large buffers (approximately 32 MB+, though it depends on
            the OS) may raise a ValueError exception
        :param input_data:
        :return:
        """
        if not self.subprocess_was_initiated:
            raise SubprocessIsNotInitiatedError

        self.wait_for_subprocess_readines(block=block)
        
        data = (True, (input_data, None))
        data = self._encode_sendable_data(data)
        # data = marshal.dumps(data)
        need_to_retry = True
        while need_to_retry:
            subprocess_disconnected_or_terminated: bool = False
            try:
                output_to_subprocess_queue = None
                if Transport.pipe == self.settings.transport:
                    output_to_subprocess_queue = self.queue_to_subprocess[0]
                elif Transport.queue == self.settings.transport:
                    output_to_subprocess_queue = self.queue_to_subprocess

                if Transport.pipe == self.settings.transport:
                    output_to_subprocess_queue.send_bytes(data)
                elif Transport.queue == self.settings.transport:
                    output_to_subprocess_queue.put(data, timeout=self.settings.subprocess_writing_timeout)
                
                need_to_retry = False
            except OSError:
                subprocess_disconnected_or_terminated = True
            except ValueError:
                subprocess_disconnected_or_terminated = True
            except Full:
                need_to_retry = block
            
            if subprocess_disconnected_or_terminated:
                self._invalidate()
                raise SubprocessTerminatedError

    def is_input_queue_is_empty(self):
        if not self.subprocess_was_initiated:
            raise SubprocessIsNotInitiatedError

        self.wait_for_subprocess_readines(block=False)
        
        result = None
        subprocess_disconnected_or_terminated: bool = False
        try:
            output_to_subprocess_queue = None
            if Transport.pipe == self.settings.transport:
                output_to_subprocess_queue = self.queue_to_subprocess[0]
            elif Transport.queue == self.settings.transport:
                output_to_subprocess_queue = self.queue_to_subprocess

            if Transport.pipe == self.settings.transport:
                result = not output_to_subprocess_queue.poll(timeout=0.0)
            elif Transport.queue == self.settings.transport:
                result = output_to_subprocess_queue.empty()

            # result = self.queue_to_subprocess.empty()
        except OSError:
            subprocess_disconnected_or_terminated = True
        except ValueError:
            subprocess_disconnected_or_terminated = True
        
        if subprocess_disconnected_or_terminated:
            self._invalidate()
            raise SubprocessTerminatedError

        return result

    def wait_for_data(self, timeout: Optional[Union[float, int]] = None):
        start_time = perf_counter()
        while not self.is_input_queue_is_empty():
            if timeout is not None:
                if (perf_counter() - start_time) >= timeout:
                    break

    def get_answer_from_subprocess(self, block=True):
        """
        If (Transport.pipe == self.settings.transport): Very large buffers (approximately 32 MB+, though it depends on
            the OS) may raise a ValueError exception
        Will raise Empty() in non-blocking mode when queue is empty
        :param block:
        :param time_out:  None - infinite; 0.0 - nonblocking; > 0.0 - timeout in seconds
        :return:
        """
        if not self.subprocess_was_initiated:
            raise SubprocessIsNotInitiatedError

        self.wait_for_subprocess_readines(block=block)
        
        subprocess_continue_working = None
        answer = None
        subprocess_disconnected_or_terminated: bool = False
        try:
            if Transport.pipe == self.settings.transport:
                input_from_subprocess_queue = self.queue_from_subprocess[0]
                if block:
                    subprocess_answer = input_from_subprocess_queue.recv_bytes()
                    subprocess_answer = self._decode_sendable_data(subprocess_answer)
                    # subprocess_answer = marshal.loads(subprocess_answer)
                    subprocess_continue_working, answer = subprocess_answer
                else:
                    if input_from_subprocess_queue.poll(timeout=0.0):
                        subprocess_answer = input_from_subprocess_queue.recv_bytes()
                        subprocess_answer = self._decode_sendable_data(subprocess_answer)
                        # subprocess_answer = marshal.loads(subprocess_answer)
                        subprocess_continue_working, answer = subprocess_answer
                    else:
                        raise Empty()
            elif Transport.queue == self.settings.transport:
                input_from_subprocess_queue = self.queue_from_subprocess
                subprocess_answer = input_from_subprocess_queue.get(block=block, timeout=self.settings.subprocess_reading_timeout)
                subprocess_answer = self._decode_sendable_data(subprocess_answer)
                # subprocess_answer = marshal.loads(subprocess_answer)
                subprocess_continue_working, answer = subprocess_answer
            
            if not subprocess_continue_working:
                subprocess_disconnected_or_terminated = True
        except OSError:
            subprocess_disconnected_or_terminated = True
        except ValueError:
            subprocess_disconnected_or_terminated = True
        
        if subprocess_disconnected_or_terminated:
            self._invalidate()
            raise SubprocessTerminatedError
        
        exception = answer[1]
        result = answer[0]
        if exception is not None:
            exception = pickle.loads(exception)
            # print(self.settings.process_name)
            # print(result)
            # print(exception)
            # print()
            # print(' <<< SUBPROCESS EXCEPTION:')
            # trace = ''
            # for line in exception[2]:
            #     trace += line
            # print(trace, file=sys.stderr)
            # print(exception[0])
            # print(exception[1])
            # print(' >>>')

            exc_type, exc_value, exc_tb = exception
            raise exc_value.with_traceback(exc_tb)
        
        return result


def _subprocess_wrapper_profile(process_data):
    printable_name = process_data.settings.process_name.replace(' ', '_') + '.prof'
    cProfile.runctx('process_data._subprocess_wrapper()', globals(), locals(), printable_name)


class ExternalPipe:
    def __init__(self):
        self.pipe = Pipe()
        self.inverted_pipe = (self.pipe[1], self.pipe[0])



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = [
    'RWOperation', 'RWLockerRequest', 'RWLockerEntity', 'RWLockerContextManager', 'FakeRWLockerContextManager', 'UnknownLockerEntity', 'RWLocker', 'get_rw_lock', 'grwl', 'aget_rw_lock', 'agrwl'
]

from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from enum import Enum
from typing import Dict, Hashable, Tuple, Union, Type, Optional, Any, List, Set
from cengal.time_management.repeat_for_a_time import Tracer
from cengal.code_flow_control.smart_values.versions.v_1 import ValueExistence
from async_generator import asynccontextmanager, async_generator, yield_
import asyncio


class RWOperation(Enum):
    read = 0
    write = 1


class RWLockerRequest(ServiceRequest):
    def register(self, entity_id: Hashable, max_writers_in_progress: int, max_readers_in_progress: int, recursive: bool = True) -> 'RWLockerContextManager':
        return self._save(0, entity_id, max_writers_in_progress, max_readers_in_progress, recursive)

    def deregister(self, entity_id: Hashable, safe: bool = True) -> bool:
        return self._save(1, entity_id, safe)
    
    def wait_for_write(self, entity_id: Hashable) -> None:
        return self._save(2, entity_id)
    
    def wait_for_read(self, entity_id: Hashable) -> None:
        return self._save(3, entity_id)


class RWLockerEntity:
    def __init__(self, entity_id: Hashable, max_writers_in_progress: int, max_readers_in_progress: int, recursive: bool, service: Type[Service]):
        self.entity_id: Hashable = entity_id
        self.recursive: bool = recursive
        self.service = service
        self.writers_pending: int = 0
        self.writers_in_progress_dict: Dict[CoroID, int] = dict()
        self.writers_in_progress: int = 0
        self.max_writers_in_progress: int = max_writers_in_progress  # Must be edited directly from coroutine. In order to eliminate new writers arrived during the end of the current loop iteration
        self.readers_pending: int = 0
        self.readers_in_progress_dict: Dict[CoroID, int] = dict()
        self.readers_in_progress: int = 0
        self.max_readers_in_progress: int = max_readers_in_progress  # Must be edited directly from coroutine. In order to eliminate new readers arrived during the end of the current loop iteration
        self.last_operation: RWOperation = RWOperation.read  # Default is 'RWOperation.read' in order to force 'write' as a first operation among several first concurent operations
        self.waiting_coroutines: Set[CoroID] = set()
        self.related_coroutines: Set[CoroID] = set()

    def check_writers_in_progress_boundaries(self, coro_id: Optional[CoroID] = None) -> bool:
        if self.recursive:
            if 0 > self.max_writers_in_progress:
                return True
            else:
                if coro_id in self.writers_in_progress_dict:
                    return True
                else:
                    return self.writers_in_progress < self.max_writers_in_progress
        else:
            if 0 > self.max_writers_in_progress:
                return True
            else:
                return self.writers_in_progress < self.max_writers_in_progress
    
    def increase_writers_in_progress(self, coro_id: Optional[CoroID] = None):
        if self.recursive:
            if coro_id not in self.writers_in_progress_dict:
                self.writers_in_progress_dict[coro_id] = 0
                self.writers_in_progress += 1
            
            self.writers_in_progress_dict[coro_id] += 1
        else:
            self.writers_in_progress += 1
    
    def decrease_writers_in_progress(self, coro_id: Optional[CoroID] = None):
        if self.recursive:
            self.writers_in_progress_dict[coro_id] -= 1
            if 0 >= self.writers_in_progress_dict[coro_id]:
                del self.writers_in_progress_dict[coro_id]
                self.writers_in_progress -= 1
        else:
            self.writers_in_progress -= 1

    def check_readers_in_progress_boundaries(self, coro_id: Optional[CoroID] = None) -> bool:
        if self.recursive:
            if 0 > self.max_readers_in_progress:
                return True
            else:
                if coro_id in self.readers_in_progress_dict:
                    return True
                else:
                    return self.readers_in_progress < self.max_readers_in_progress
        else:
            if 0 > self.max_readers_in_progress:
                return True
            else:
                return self.readers_in_progress < self.max_readers_in_progress
    
    def increase_readers_in_progress(self, coro_id: Optional[CoroID] = None):
        if self.recursive:
            if coro_id not in self.readers_in_progress_dict:
                self.readers_in_progress_dict[coro_id] = 0
                self.readers_in_progress += 1
            
            self.readers_in_progress_dict[coro_id] += 1
        else:
            self.readers_in_progress += 1
    
    def decrease_readers_in_progress(self, coro_id: Optional[CoroID] = None):
        if self.recursive:
            self.readers_in_progress_dict[coro_id] -= 1
            if 0 >= self.readers_in_progress_dict[coro_id]:
                del self.readers_in_progress_dict[coro_id]
                self.readers_in_progress -= 1
        else:
            self.readers_in_progress -= 1

    def try_write_lock(self, coro_id: Optional[CoroID] = None, apply: bool = True) -> bool:
        need_to_try_later = False
        if self.readers_in_progress or self.readers_pending:
            if (not self.readers_in_progress) and self.readers_pending and (RWOperation.read == self.last_operation) and self.check_writers_in_progress_boundaries(coro_id):
                if apply:
                    self.increase_writers_in_progress(coro_id)
                    self.last_operation = RWOperation.write
            else:
                need_to_try_later = True
        else:
            if self.check_writers_in_progress_boundaries(coro_id):
                if apply:
                    self.increase_writers_in_progress(coro_id)
                    self.last_operation = RWOperation.write
            else:
                need_to_try_later = True
        
        return need_to_try_later

    def try_read_lock(self, coro_id: Optional[CoroID] = None, apply: bool = True) -> bool:
        need_to_try_later = False
        if self.writers_in_progress or self.writers_pending:
            if (not self.writers_in_progress) and self.writers_pending and (RWOperation.write == self.last_operation) and self.check_readers_in_progress_boundaries(coro_id):
                if apply:
                    self.increase_readers_in_progress(coro_id)
                    self.last_operation = RWOperation.read
            else:
                need_to_try_later = True
        else:
            if self.check_readers_in_progress_boundaries(coro_id):
                if apply:
                    self.increase_readers_in_progress(coro_id)
                    self.last_operation = RWOperation.read
            else:
                need_to_try_later = True
        
        return need_to_try_later
    
    def test_remove(self) -> bool:
        need_to_try_later = self.waiting_coroutines
        return need_to_try_later


class RWLockerContextManagerBase:
    def __init__(self, core: RWLockerEntity) -> None:
        self.core: RWLockerEntity = core
        self.current_context_operation: Optional[RWOperation] = None
    
    def lockable(self, operation: Optional[RWOperation] = None) -> bool:
        if operation is None:
            operation = RWOperation.read
        
        if RWOperation.write == operation:
            need_service_assistance = self.core.try_write_lock(self._interface.coro_id, False)
        else:
            need_service_assistance = self.core.try_read_lock(self._interface.coro_id, False)
        
        return not need_service_assistance
    
    def change_max_boundaries(self, max_writers_in_progress: int, max_readers_in_progress: int):
        self.core.max_writers_in_progress = max_writers_in_progress
        self.core.max_readers_in_progress = max_readers_in_progress
        
    def __call__(self, operation: Optional[RWOperation] = None):
        if operation is None:
            operation = RWOperation.read
        
        self.current_context_operation = operation
        return self
    

class RWLockerContextManager(RWLockerContextManagerBase):
    def __init__(self, core: RWLockerEntity, interface: Interface):
        super().__init__(core)
        self._interface = interface
    
    def __enter__(self):
        if self.current_context_operation is None:
            self.current_context_operation = RWOperation.read
        
        if RWOperation.write == self.current_context_operation:
            need_service_assistance = self.core.try_write_lock(self._interface.coro_id)
            if need_service_assistance:
                self.core.writers_pending += 1
                self._interface(self.core.service, RWLockerRequest().wait_for_write(self.core.entity_id))
        else:
            need_service_assistance = self.core.try_read_lock(self._interface.coro_id)
            if need_service_assistance:
                self.core.readers_pending += 1
                self._interface(self.core.service, RWLockerRequest().wait_for_read(self.core.entity_id))
        
        return self
    
    def __exit__(self, type, value: Exception, traceback):
        if RWOperation.write == self.current_context_operation:
            self.core.decrease_writers_in_progress(self._interface)
        else:
            self.core.decrease_readers_in_progress(self._interface)
        
        self.current_context_operation = None

    async def __aenter__(self):
        if self.current_context_operation is None:
            self.current_context_operation = RWOperation.read
        
        if RWOperation.write == self.current_context_operation:
            need_service_assistance = self.core.try_write_lock(self._interface.coro_id)
            if need_service_assistance:
                self.core.writers_pending += 1
                await self._interface(self.core.service, RWLockerRequest().wait_for_write(self.core.entity_id))
        else:
            need_service_assistance = self.core.try_read_lock(self._interface.coro_id)
            if need_service_assistance:
                self.core.readers_pending += 1
                await self._interface(self.core.service, RWLockerRequest().wait_for_read(self.core.entity_id))        
        
        return self

    async def __aexit__(self, type, value, traceback):
        if RWOperation.write == self.current_context_operation:
            self.core.decrease_writers_in_progress(self._interface)
        else:
            self.core.decrease_readers_in_progress(self._interface)
        
        self.current_context_operation = None


class FakeRWLockerContextManager(RWLockerContextManagerBase):
    def __init__(self, core: RWLockerEntity):
        super().__init__(core)
    
    def __enter__(self):
        if self.current_context_operation is None:
            self.current_context_operation = RWOperation.read
        
        if RWOperation.write == self.current_context_operation:
            self.core.increase_writers_in_progress()
            self.core.last_operation = RWOperation.write
        else:
            self.core.increase_readers_in_progress()
            self.core.last_operation = RWOperation.read
        
        return self
    
    def __exit__(self, type, value: Exception, traceback):
        if RWOperation.write == self.current_context_operation:
            self.core.decrease_writers_in_progress()
        else:
            self.core.decrease_readers_in_progress()
        
        self.current_context_operation = None

    async def __aenter__(self):
        if self.current_context_operation is None:
            self.current_context_operation = RWOperation.read
        
        if RWOperation.write == self.current_context_operation:
            self.core.increase_writers_in_progress()
            self.core.last_operation = RWOperation.write
        else:
            self.core.increase_readers_in_progress()
            self.core.last_operation = RWOperation.read
        
        return self

    async def __aexit__(self, type, value, traceback):
        if RWOperation.write == self.current_context_operation:
            self.core.decrease_writers_in_progress()
        else:
            self.core.decrease_readers_in_progress()
        
        self.current_context_operation = None


class UnknownLockerEntity(Exception):
    pass


class RWLocker(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(RWLocker, self).__init__(loop)

        # loop.add_global_on_coro_del_handler(self._on_coro_del_handler_global)  # Todo: switch to local coro del handler

        self._request_workers = {
            0: self._on_register,
            1: self._on_deregister,
            2: self._on_wait_for_write,
            3: self._on_wait_for_read,
        }
        
        self.locker_entities: Dict[Hashable, RWLockerEntity] = dict()
        self.entities_by_coroutine: Dict[CoroID, Set[Hashable]] = dict()
        
        self.remove_entity_requests: Dict[CoroID, Hashable] = dict()
        self.waiting_for_write_requests: Dict[CoroID, Hashable] = dict()
        self.waiting_for_read_requests: Dict[CoroID, Hashable] = dict()

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'locker entities num': len(self.locker_entities),
            'affected coroutines num': len(self.entities_by_coroutine),
            'waiting for write requests num': len(self.waiting_for_write_requests),
            'waiting_for_read_requests num': len(self.waiting_for_read_requests),
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[RWLockerRequest]=None
                                                         ) -> ServiceProcessingResponse:
        if request is not None:
            return self.resolve_request(request)
        return True, None, None

    def full_processing_iteration(self):
        # entities_waiting_for_remove
        processed_coro_ids: Set[CoroID] = set()
        for coro_id, entity_id in self.remove_entity_requests.items():
            if entity_id not in self.locker_entities:
                self.register_response(coro_id, False, None)
                processed_coro_ids.add(coro_id)
                continue
            
            entity = self.locker_entities[entity_id]
            need_to_try_later = entity.test_remove()
            if need_to_try_later:
                continue
            else:
                del self.locker_entities[entity_id]
                processed_coro_ids.add(coro_id)
        
        for coro_id in processed_coro_ids:
            del self.remove_entity_requests[coro_id]
        
        # entities_waiting_for_write
        processed_coro_ids: Set[CoroID] = set()
        for coro_id, entity_id in self.waiting_for_write_requests.items():
            if entity_id not in self.locker_entities:
                self.register_response(coro_id, None, UnknownLockerEntity)
                processed_coro_ids.add(coro_id)
                continue
            
            entity = self.locker_entities[entity_id]
            need_to_try_later = entity.try_write_lock(coro_id)
            if need_to_try_later:
                continue
            else:
                entity.writers_pending -= 1
                if coro_id in entity.waiting_coroutines:
                    entity.waiting_coroutines.remove(coro_id)
                
                self.register_response(coro_id, None, None)
                processed_coro_ids.add(coro_id)
        
        for coro_id in processed_coro_ids:
            del self.waiting_for_write_requests[coro_id]
                    
        # entities_waiting_for_read
        processed_coro_ids: Set[CoroID] = set()
        for coro_id, entity_id in self.waiting_for_read_requests.items():
            if entity_id not in self.locker_entities:
                self.register_response(coro_id, None, UnknownLockerEntity)
                processed_coro_ids.add(coro_id)
                continue
            
            entity = self.locker_entities[entity_id]
            need_to_try_later = entity.try_read_lock(coro_id)
            if need_to_try_later:
                continue
            else:
                entity.readers_pending -= 1
                if coro_id in entity.waiting_coroutines:
                    entity.waiting_coroutines.remove(coro_id)
                
                self.register_response(coro_id, None, None)
                processed_coro_ids.add(coro_id)
        
        for coro_id in processed_coro_ids:
            del self.waiting_for_read_requests[coro_id]
        
        # general
        if not (self.remove_entity_requests or self.waiting_for_write_requests or self.waiting_for_read_requests):
            self.make_dead()

    def in_work(self) -> bool:
        result: bool = bool(self.remove_entity_requests) or bool(self.waiting_for_write_requests) or bool(self.waiting_for_read_requests)
        return self.thrifty_in_work(result)
    
    def get_locker_entity(self, entity_id: Hashable):
        return self.locker_entities.get(entity_id)

    def _on_register(self, entity_id: Hashable, max_writers_in_progress: int, max_readers_in_progress: int, recursive: bool = True) -> ServiceProcessingResponse:
        if entity_id not in self.locker_entities:
            self.locker_entities[entity_id] = RWLockerEntity(entity_id, max_writers_in_progress, max_readers_in_progress, recursive, type(self))
        entity: RWLockerEntity = self.locker_entities[entity_id]
        
        coro_id = self.current_caller_coro_info.coro.coro_id
        entity.related_coroutines.add(coro_id)
        if coro_id not in self.entities_by_coroutine:
            self.entities_by_coroutine[coro_id] = set()
        
        self.entities_by_coroutine[coro_id].add(entity_id)
        self.current_caller_coro_info.coro.add_on_coro_del_handler(self._on_coro_del_handler)
        context_manager: RWLockerContextManager = RWLockerContextManager(entity, self.current_caller_coro_info.coro.interface)
        return True, context_manager, None

    def _on_deregister(self, entity_id: Hashable, safe: bool = True) -> ServiceProcessingResponse:
        result = None
        if safe:
            self.remove_entity_requests[self.current_caller_coro_info.coro.coro_id] = entity_id
            self.current_caller_coro_info.coro.add_on_coro_del_handler(self._on_coro_del_handler)
            self.make_live()
            return False, None, None
        else:
            if entity_id in self.locker_entities:
                entity = self.locker_entities[entity_id]
                del self.locker_entities[entity_id]
                for related_coro_id in entity.related_coroutines:
                    coroutine_entities = self.entities_by_coroutine[related_coro_id]
                    if entity_id in coroutine_entities:
                        coroutine_entities.remove(entity_id)
                
                result = True
            else:
                result = False
            
            return True, result, None

    def _on_wait_for_write(self, entity_id: Hashable) -> ServiceProcessingResponse:
        if entity_id not in self.locker_entities:
            return True, None, UnknownLockerEntity()
        
        coro_id = self.current_caller_coro_info.coro.coro_id
        self.waiting_for_write_requests[coro_id] = entity_id
        self.current_caller_coro_info.coro.add_on_coro_del_handler(self._on_coro_del_handler)
        entity = self.locker_entities[entity_id]
        entity.waiting_coroutines.add(coro_id)
        self.make_live()
        return False, None, None

    def _on_wait_for_read(self, entity_id: Hashable) -> ServiceProcessingResponse:
        if entity_id not in self.locker_entities:
            return True, None, UnknownLockerEntity()
        
        coro_id = self.current_caller_coro_info.coro.coro_id
        self.waiting_for_read_requests[coro_id] = entity_id
        self.current_caller_coro_info.coro.add_on_coro_del_handler(self._on_coro_del_handler)
        entity = self.locker_entities[entity_id]
        entity.waiting_coroutines.add(coro_id)
        self.make_live()
        return False, None, None

    def _on_coro_del_handler_global(self, coro: CoroWrapperBase) -> bool:
        coro_id = coro.coro_id
        if coro_id in self.entities_by_coroutine:
            entities = self.entities_by_coroutine[coro_id]
            del self.entities_by_coroutine[coro_id]
            for entity_id in entities:
                if entity_id in self.locker_entities:
                    entity = self.locker_entities[entity_id]
                    if coro_id in entity.related_coroutines:
                        entity.related_coroutines.remove(coro_id)
                    
                    if coro_id in entity.waiting_coroutines:
                        entity.waiting_coroutines.remove(coro_id)
        
        if coro_id in self.remove_entity_requests:
            del self.remove_entity_requests[coro_id]
        
        if coro_id in self.waiting_for_write_requests:
            del self.waiting_for_write_requests[coro_id]
        
        if coro_id in self.waiting_for_read_requests:
            del self.waiting_for_read_requests[coro_id]

        return False

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        return self._on_coro_del_handler_global(coro)


RWLockerRequest.default_service_type = RWLocker


def get_rw_lock(entity_id: Hashable, max_writers_in_progress: int, max_readers_in_progress: int, recursive: bool = True) -> Union[RWLockerContextManager, FakeRWLockerContextManager]:
    loop = CoroScheduler.current_loop()
    if loop is None:
        return FakeRWLockerContextManager(RWLockerEntity(entity_id, max_writers_in_progress, max_readers_in_progress, recursive, RWLocker))  # running not from inside the loop

    interface = loop.current_interface()
    if interface is None:
        return FakeRWLockerContextManager(RWLockerEntity(entity_id, max_writers_in_progress, max_readers_in_progress, recursive, RWLocker))  # running from Service

    locker_entity = interface._loop.get_service_instance(RWLocker).get_locker_entity(entity_id)
    if locker_entity is None:
        lock = interface(RWLocker, RWLockerRequest().register(entity_id, max_writers_in_progress, max_readers_in_progress, recursive))
    else:
        lock = RWLockerContextManager(locker_entity, interface)
    
    return lock


grwl = get_rw_lock


async def aget_rw_lock(entity_id: Hashable, max_writers_in_progress: int, max_readers_in_progress: int, recursive: bool = True) -> Union[RWLockerContextManager, FakeRWLockerContextManager]:
    loop = CoroScheduler.current_loop()
    if loop is None:
        return FakeRWLockerContextManager(RWLockerEntity(entity_id, max_writers_in_progress, max_readers_in_progress, recursive, RWLocker))  # running not from inside the loop

    interface = loop.current_interface()
    if interface is None:
        return FakeRWLockerContextManager(RWLockerEntity(entity_id, max_writers_in_progress, max_readers_in_progress, recursive, RWLocker))  # running from Service

    locker_entity = interface._loop.get_service_instance(RWLocker).get_locker_entity(entity_id)
    if locker_entity is None:
        lock = await interface(RWLocker, RWLockerRequest().register(entity_id, max_writers_in_progress, max_readers_in_progress, recursive))
    else:
        lock = RWLockerContextManager(locker_entity, interface)
    
    return lock


agrwl = aget_rw_lock



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['DirNameOrPath', 'BaseDirID', 'DirTypeMapping', 'DirTypeMappingItem', 'norm_dir_name_or_path', 'AppDirPathBase']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import os
from cengal.file_system.directory_manager import ensure_dir as ensure_dir_exists
from .app_directory_types import AppDirectoryType
from typing import Union, Sequence, Any, Optional, Tuple, Dict
from functools import lru_cache


DirNameOrPath = Union[str, Sequence[str], None]
BaseDirID = Any
DirTypeMapping = Dict[AppDirectoryType, Tuple[BaseDirID, Optional[DirNameOrPath], Optional[DirNameOrPath]]]
DirTypeMappingItem = Tuple[BaseDirID, Optional[DirNameOrPath], Optional[DirNameOrPath]]


def norm_dir_name_or_path(dir_name_or_path: DirNameOrPath) -> Tuple[str, ...]:
    if dir_name_or_path is None:
        return ()
    elif isinstance(dir_name_or_path, str):
        return (dir_name_or_path,)
    elif isinstance(dir_name_or_path, (list, tuple)):
        # return tuple(dir_name_or_path)
        return dir_name_or_path
    else:
        raise TypeError('dir_name_or_path must be str, list or tuple')


class AppDirPathBase:
    def __init__(self, max_cache_size: Optional[int] = None) -> None:
        if max_cache_size is None:
            max_cache_size = len(AppDirectoryType) * 2 * 2 * 10  # 10 - is a number of applications; 2 - is `with_structure` [True/False]; 2 - is `ensure_dir` [True/False]

        self._max_cache_size = max_cache_size
        self._cached = self.__call__
        self.set_cache_size(max_cache_size)
    
    def get_cache_size(self) -> int:
        return self._max_cache_size
    
    def set_cache_size(self, max_cache_size: int) -> None:
        self._max_cache_size = max_cache_size
        del self._cached
        self._cached = lru_cache(maxsize=max_cache_size, typed=True)(self.__call__)
    
    def dir_type_mapping(self, dir_type: AppDirectoryType) -> DirTypeMappingItem:
        raise NotImplementedError
    
    def base_dir_id_to_path(self, base_dir_id: BaseDirID) -> str:
        raise NotImplementedError

    def __call__(self, dir_type: AppDirectoryType, app_name_or_path: Optional[DirNameOrPath] = None, with_structure: bool = True, ensure_dir: bool = True) -> str:
        if app_name_or_path is not None:
            app_name_or_path = str(app_name_or_path)
        
        if AppDirectoryType.user_profile_data == dir_type:
            if app_name_or_path is not None:
                app_name_or_path = '.' + os.path.join(*norm_dir_name_or_path(app_name_or_path))

        mapping: DirTypeMappingItem = self.dir_type_mapping(dir_type)
        base_dir_path = self.base_dir_id_to_path(mapping[0])
        result_list = [base_dir_path]
        result_list.extend(norm_dir_name_or_path(mapping[1]))
        if app_name_or_path is not None:
            result_list.extend(norm_dir_name_or_path(app_name_or_path))
            if with_structure:
                result_list.extend(norm_dir_name_or_path(mapping[2]))
        
        result_path = os.path.normpath(os.path.join(*result_list))
        if ensure_dir:
            ensure_dir_exists(result_path)
        
        return result_path

    def cached(self, dir_type: AppDirectoryType, app_name_or_path: Optional[DirNameOrPath] = None, with_structure: bool = True, ensure_dir: bool = True) -> str:
        return self._cached(dir_type, app_name_or_path, with_structure, ensure_dir)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import errno
import copy
import enum
from contextlib import contextmanager

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class LoopIsAlreadyBegun(Exception):
    """
    You can not run NetIOUserApi.start() if it was already started (and still running).
    """
    pass


class WrongConnectionType(Exception):
    """
    You cannot run NetIOUserApi.make_connection() for ConnectionType.active_accepted connection. This kind of
    connections are made only from (and by) inside of IO loop and logic itself.
    """
    pass


class CanNotMakeConnection(Exception):
    """
    Currently not used. If there will be some exception on connect() call - it will be raised without any changes.
    """
    pass


class ConnectionType(enum.Enum):
    passive = 0  # passive socket (bind())
    active_accepted = 1  # active accepted socket (accept())
    active_connected = 2  # active connected socket (connect())


class ConnectionState(enum.Enum):
    not_connected_yet = 0  # socket is not in connection process
    waiting_for_connection = 1  # socket is in connection process (async connection is delayed)
    connected = 2  # socket is successfully connected
    worker_fault = 3  # there was unhandled exception from one of the WorkerBase callbacks
    io_fault = 4  # there was some IO trouble
    waiting_for_disconnection = 5  # connection was marked as "should be closed" but was not closed yet
    disconnected = 6  # socket is closed


class ConnectionInfo:
    def __init__(self,
                 worker_obj,
                 connection_type: ConnectionType,
                 socket_address=None,
                 socket_family=socket.AF_INET,
                 socket_type=socket.SOCK_STREAM,
                 socket_protocol=0,
                 socket_fileno=None,
                 backlog=0):
        """
        :param worker_obj: constructed worker object (see WorkerBase for more info). If this is a passive
            connection - it (worker_obj) will be inherited by the descendant active_accepted connections
            by copy.copy() call (see WorkerBase.__copy__() method for more info)
        :param connection_type: see ConnectionType() description
        :param socket_address:  see socket.bind()/socket.connect() docs
        :param socket_family: see socket.socket() docs
        :param socket_type: see socket.socket() docs
        :param socket_protocol: see socket.socket() docs
        :param socket_fileno: see socket.socket() docs
        :param backlog: see socket.listen() docs
        """
        self.worker_obj = worker_obj
        self.connection_type = connection_type
        self.socket_address = socket_address
        self.socket_family = socket_family
        self.socket_type = socket_type
        self.socket_protocol = socket_protocol
        self.socket_fileno = socket_fileno
        self.backlog = backlog


class Connection:
    """
    Connection class. Usually created by IO loop or by IO API. But you can also create it by yourself
    """
    def __init__(self,
                 connection_id,
                 connection_info: ConnectionInfo,
                 connection_and_address_pair: tuple,
                 connection_state: ConnectionState,
                 connection_name=None,
                 ):
        """
        :param connection_id: unique connection identificator (unique within the IO object). You may use some random
            GUID if you are creating connection by your self.
        :param connection_info: new connection will be created using information provided in connection_info object.
            See ConnectionInfo() for more information
        :param connection_and_address_pair: (conn, address) tuple where conn is connected socket (or it can be socket
            is in the process of connection. But only if it was made by IO loop.).
        :param connection_state: see ConnectionState for more information
        :param connection_name: name of the connection (if it was provided)
        """
        self.connection_id = connection_id
        self.connection_info = connection_info
        self.conn, self.address = connection_and_address_pair
        self.connection_state = connection_state
        self.connection_name = connection_name
        self.worker_obj = connection_info.worker_obj
        self.read_data = b''  # already read data
        self.must_be_written_data = memoryview(b'')  # this data should be written
        self.force_write_call = False

    def add_must_be_written_data(self, data):
        """
        Use this method to add data to output buffers
        :param data: some new output data to be send through this connection
        :return:
        """
        self.must_be_written_data = memoryview(bytes(self.must_be_written_data) + data)


class NetIOUserApi:
    """
    You may rely and use freely use methods of this base class from inside your program or from inside your worker
    (WorkerBase).
    """
    def __init__(self):
        super().__init__()
        self.all_connections = set()
        self.passive_connections = set()

        self.connection_by_id = dict()
        self.connection_by_name = dict()
        self.connection_by_fileno = dict()

    def start(self, destroy_on_finish=True):
        """
        Will start IO loop
        :param destroy_on_finish: if True - destroy() will be called from inside of this method
        :return:
        """
        raise NotImplementedError()

    def stop(self):
        """
        Will initiate IO loop stop process
        :return:
        """
        raise NotImplementedError()

    def make_connection(self, connection_info: ConnectionInfo = None, name=None)->Connection:
        """
        Will create connection from given connection_info object. Than connection will be established. Immediate or
        delayed - depends on the connection type:
        - ConnectionType.passive - immediate;
        - ConnectionType.active_connected - delayed.
        In both cases WorkerBase.on_connect will be called immediately after connection will be successfully
        established (IF it will be successfully established).
        :param connection_info: new connection will be created using information provided in connection_info object.
            See ConnectionInfo() for more information
        :param name: name of the connection. If you'll provide it - you will be able to find this connection in
            NetIOUserApi.connection_by_name dictionary by it's name.
        :return:
        """
        raise NotImplementedError()

    def add_connection(self, connection: Connection):
        """
        Will register already established connection. You need to use this method for example if you have already
        connected socket
        :param connection:
        :return:
        """
        raise NotImplementedError()

    def remove_connection(self, connection: Connection):
        """
        Will close and remove connection
        :param connection:
        :return:
        """
        raise NotImplementedError()

    def check_is_connection_need_to_sent_data(self, connection: Connection):
        """
        Will check connection to output data presence. It is automatically called after EACH WorkerBase callback call
        by the IO loop. But if you are filling other connection's output buffer - you'll need to make this call for that
        connection by your self.
        :param connection:
        :return:
        """
        raise NotImplementedError()


class NetIOCallbacks:
    """
    Callbacks from this class will be called from inside (and by) IOMethodBase loop.
    """
    def __init__(self):
        super().__init__()

    def on_accept_connection(self, connection: Connection):
        raise NotImplementedError()

    def on_connected(self, connection: Connection):
        raise NotImplementedError()

    def on_read(self, connection: Connection):
        raise NotImplementedError()

    def on_write(self, connection: Connection):
        raise NotImplementedError()

    def on_close(self, connection: Connection):
        raise NotImplementedError()


class NetIOBase(NetIOUserApi, NetIOCallbacks):
    """
    Base class for any IO implementation (Linux, BSD, Windows, cross platform, etc.).
    """
    def __init__(self, transport):
        """

        :param transport: class of the desired IOMethod. Instance (object) will be created by NetIOBase itself
        """
        super().__init__()
        self.method = transport(self)

    def destroy(self):
        raise NotImplementedError()


class WorkerBase:
    """
    Base class for all workers.
    on_* callbacks will be called by the IO loop.

    General info:
    You can read input data from self.connection at any time (see "Caution" section of __init__ doc string) from any
    callback.
    You can write output data (to be send) to self.connection at any time (see "Caution") from any callback.
    """
    def __init__(self, api: NetIOUserApi=None, connection: Connection=None):
        """
        Caution:
        Please do not rely on self.api and self.connection inside of your __init__ constructor: it is guaranteed that
        they will be set before any callback call, but not at the construction process.

        :param api: link to the constructed network io object
        :param connection: link to the constructed connection object
        """
        self.api = api
        self.connection = connection

    def on_connect(self):
        """
        Will be called after connection successfully established
        :return:
        """
        pass

    def on_read(self):
        """
        Will be called if there is some NEW data in the connection input buffer
        :return:
        """
        pass

    def on_no_more_data_to_write(self):
        """
        Will be called after all data is sent.
        Normally it will be called once (one single shot after each portion of out data is sent).
        If you'll set self.connection.force_write_call to True state - this callback may be called continuously
        (but not guaranteed: it depends of used IOMethod implementation)
        :return:
        """
        pass

    def on_connection_lost(self):
        """
        Will be called AFTER connection socket was actually closed and removed from IOMethod checking list.
        At this time, self.connection.connection_state is already set to ConnectionState.disconnected.
        :return:
        """
        pass

    def __copy__(self):
        """
        This method SHOULD be implemented. It should create a new instance and copy some global (shared) data from
        current object to that new instance. It will be called when new peer is connected to existing passive connection
        So this is the way you may use to give all new connection some links to some global data by worker object of
        the passive connection replication process.
        :return:
        """
        raise NotImplementedError()


@contextmanager
def net_io(net_io_obj: NetIOBase):
    """
    Context manager.
    Usage:

    main_io = NetIOBase()
    with(net_io(main_io)) as io:
        print('Preparing connections')
        connection1 = io.make_connection(...)
        connection2 = io.make_connection(...)
        k = c + 12
        ...
        connectionN = io.make_connection(...)
        print('Starting IO loop')
    print('IO loop was finished properly')


    :param net_io_obj: constructed IO instance (object)
    :return:
    """
    try:
        yield net_io_obj
        net_io_obj.start(destroy_on_finish=False)
    finally:
        net_io_obj.destroy()


class IOMethodBase:
    """
    Base class for all IOMethod implementation (select, epoll, overlapped io, kqueue, etc.)
    All his methods are called by the NetIOBase instance.
    """
    def __init__(self, interface: NetIOBase):
        self.interface = interface
        self.should_be_closed = set()
        pass

    def loop_iteration(self, timeout=None):
        """
        Single IO loop iteration.
        This method holds all IOMethod logic.
        :return:
        """
        raise NotImplementedError()

    def destroy(self):
        """
        Initiates destruction process
        :return:
        """
        raise NotImplementedError()

    def set__can_read(self, conn: socket.socket, state=True):
        """
        Will allow (True) or disallow (False) "socket available to read" checks for socket
        :param conn: target socket
        :param state: True - allow; False - disallow
        :return:
        """
        raise NotImplementedError()

    def set__need_write(self, conn: socket.socket, state=True):
        """
        Will allow (True) or disallow (False) "socket available to write" checks for socket
        :param conn: target socket
        :param state: True - allow; False - disallow
        :return:
        """
        raise NotImplementedError()

    def set__should_be_closed(self, conn: socket.socket):
        """
        Mark socket as "should be closed"
        :param conn: target socket
        :return:
        """
        raise NotImplementedError()

    def add_connection(self, conn: socket.socket):
        """
        Will add socket to the internal connections list
        :param conn: target socket
        :return:
        """
        raise NotImplementedError()

    def remove_connection(self, conn: socket.socket):
        """
        Will remove socket from the internal connections list
        :param conn: target socket
        :return:
        """
        raise NotImplementedError()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['TextTranslationDictionary', 'TextEntityId', 'TranslatableText', 'TranslationLanguageId', 'TextTranslatorError', 'TextTranslator', 'TranslationLanguageMapper', 'TranslationLanguageChooser', 'TextTranslationReapplier', 'CoroPriority', 'tt', 'TranslationWorker']


from collections.abc import Mapping
from typing import Hashable, Dict, Union, Optional, Callable
from cengal.data_manipulation.serialization import *
from cengal.code_flow_control.call_history_reapplier import *
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import *
from uuid import uuid4
import sys


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


TranslationLanguageId = Hashable
TextEntityId = Hashable
TranslatableText = str
TextTranslationDictionary = Dict
'''
In Python:
text_translation_dictionary = {
    '{TranslatableText}': {
        'default': {  // Can be empty. "Variants" field must not be empty in this case.
            '{TranslationLanguageId}': '{TranslatableText}',
            '{TranslationLanguageId}': '{TranslatableText}',
            ...
        },
        'variants': {  // Can be empty. "Default" field must not be empty in this case.
            '{TextEntityId}': {
                '{TranslationLanguageId}': '{TranslatableText}',
                '{TranslationLanguageId}': '{TranslatableText}',
                ...
            },
            '{TextEntityId}': {
                '{TranslationLanguageId}': '{TranslatableText}',
                '{TranslationLanguageId}': '{TranslatableText}',
                ...
            },
            ...
        }
    }
}

In JSON:
{
    'type': 'Cengal.TextTranslationDictionary',
    'version': '1.0.0'
    'text_translation_list': [
        {
            'text': '{TranslatableText}',
            'translations': {
                'default': {  // Can be empty. "Variants" field must not be empty in this case.
                    '{TranslationLanguageId}': '{TranslatableText}',
                    '{TranslationLanguageId}': '{TranslatableText}',
                    ...
                },
                'variants': {  // Can be empty. "Default" field must not be empty in this case.
                    '{TextEntityId}': {
                        '{TranslationLanguageId}': '{TranslatableText}',
                        '{TranslationLanguageId}': '{TranslatableText}',
                        ...
                    },
                    '{TextEntityId}': {
                        '{TranslationLanguageId}': '{TranslatableText}',
                        '{TranslationLanguageId}': '{TranslatableText}',
                        ...
                    },
                    ...
                }
            }
        }
    ]
}
'''
TranslationLangToLangMap = Dict[TranslationLanguageId, TranslationLanguageId]
TranslationWorker = Callable[['TranslatableText'], None]


class TextTranslatorError(Exception):
    pass

class TextTranslator:
    @classmethod
    def from_json(cls, json_data: Union[bytes, str], encoding: Optional[str]=None):
        serializer = best_serializer_for_standard_data((DataFormats.json,
                                 Tags.decode_str_as_str,
                                 Tags.decode_list_as_list,
                                 Tags.superficial,
                                 Tags.current_platform),
                                TestDataType.deep_large,
                                0.1)
        encoding = encoding or 'utf-8'
        if isinstance(json_data, bytes):
            json_data = json_data.decode(encoding=encoding)
        decoded_data: Dict = serializer.loads(json_data)
        if not isinstance(decoded_data, Mapping):
            raise TextTranslatorError('Wrong json data: root must be a dict')
        if 'Cengal.TextTranslationDictionary' != decoded_data.get('type'):
            raise TextTranslatorError('Wrong json data: lack of "type" field or a "type" field value mismatch')
        try:
            text_translation_dictionary: TextTranslationDictionary = dict()
            text_translation_list = decoded_data['text_translation_list']
            for item in text_translation_list:
                text_translation_dictionary[item['text']] = item['translations']
            return cls(text_translation_dictionary, decoded_data)
        except (KeyError, TypeError):
            raise TextTranslatorError('Wrong json data or other parsing error')
    
    def __init__(self, dictionary: TextTranslationDictionary, decoded_data: Optional[Dict]=None):
        self.dictionary = dictionary
        self.decoded_data = decoded_data
    
    def __call__(self, language: TranslationLanguageId, text: TranslatableText, entity_id: Optional[TextEntityId]=None) -> TranslatableText:
        try:
            translations = self.dictionary[text]
            if entity_id is None:
                variant = translations['default']
            else:
                variant = translations['variants'][entity_id]
            return variant[language]
        except KeyError:
            return text


class TranslationLanguageMapper:
    def __init__(self, lang_2_lang: TranslationLangToLangMap, default_lang: TranslationLanguageId):
        self.lang_2_lang: TranslationLangToLangMap = lang_2_lang
        self.default_lang: TranslationLanguageId = default_lang
    
    def __call__(self, lang: TranslationLanguageId):
        return self.lang_2_lang.get(lang, None) or self.default_lang


class TranslationLanguageChooser:
    def __init__(self, text_translator: TextTranslator, 
                 translation_language_mapper: TranslationLanguageMapper,
                 coro_scheduler: Optional[CoroSchedulerType]=None):
        self._end_lang: Optional[TranslationLanguageId] = None
        self._lang: Optional[TranslationLanguageId] = None
        self.text_translator: TextTranslator = text_translator
        self.translation_language_mapper: TranslationLanguageMapper = translation_language_mapper
        self.coro_scheduler: CoroSchedulerType = coro_scheduler or CoroScheduler.current_loop()
        self.translation_language_changed_event: str = str(uuid4())
    
    @property
    def lang(self) -> TranslationLanguageId:
        return self._lang
    
    @lang.setter
    def lang(self, language: TranslationLanguageId):
        self._lang = language
        self._end_lang = self.translation_language_mapper(self._lang)
        
        # def raise_translation_language_changed_event(
        #     interface: Interface,
        #     translation_language_changed_event: str,
        #     lang: TranslationLanguageId,
        #     end_lang: TranslationLanguageId
        #     ):
        #     with log_uncatched_exception():
        #         print('raise_translation_language_changed_event - raising event...')
        #         print(interface, translation_language_changed_event, lang, end_lang)
        #         interface(Sleep, 1.0)
        #         interface(AsyncEventBus, AsyncEventBusRequest().send_event(translation_language_changed_event, (lang, end_lang), CoroPriority.low))
        #         interface(Sleep, 1.0)
        #         print('raise_translation_language_changed_event - done')
        
        # try_put_coro_to(get_interface_and_loop_with_explicit_loop(self.coro_scheduler), raise_translation_language_changed_event, 
        #         self.translation_language_changed_event, self._lang, self._end_lang)
        try_send_async_event(self.coro_scheduler, self.translation_language_changed_event, (self._lang, self._end_lang), CoroPriority.high)
    
    # @staticmethod
    # def raise_translation_language_changed_event(
    #     interface: Interface,
    #     translation_language_changed_event: str,
    #     lang: TranslationLanguageId,
    #     end_lang: TranslationLanguageId
    #     ):
    #     print('raise_translation_language_changed_event - raising event...')
    #     print(interface, translation_language_changed_event, lang, end_lang)
    #     interface(AsyncEventBus, AsyncEventBusRequest().send_event(translation_language_changed_event, (lang, end_lang), CoroPriority.low))
    #     print('raise_translation_language_changed_event - done')
    
    def set_lang(self, language: TranslationLanguageId) -> 'TranslationLanguageChooser':
        '''
        For usage with ArgsManager like
                am = ArgsManager(
                    EArgs(text_translator=TranslationLanguageChooser(
                        TextTranslator.from_json(TEXT_DICTIONARY), 
                        TranslationLanguageMapper(TRANSLATION_LANGUAGE_MAP, 'en')).set_lang('ru'))
                )

        '''
        self.lang = language
        return self
    
    @property
    def end_lang(self) -> TranslationLanguageId:
        return self._end_lang
    
    @end_lang.setter
    def end_lang(self, language: TranslationLanguageId):
        pass
    
    def __call__(self, text: TranslatableText, entity_id: Optional[TextEntityId]=None) -> TranslatableText:
        return self.text_translator(self._end_lang, text, entity_id)


class TextTranslationReapplier(CallHistoryReapplier):
    class ToBeTranslated:
        def __init__(self):
            self.text: TranslatableText = None
        
        def __call__(self, text: TranslatableText) -> 'TextTranslationReapplier.ToBeTranslated':
            self.text = text
            return self
        
    def __init__(self, text_translator: TranslationLanguageChooser, priority: CoroPriority=CoroPriority.low):
        self.text_translator = text_translator
        super().__init__(priority)
    
    def _translate_needed(self, value: Any, entity_id: Optional[TextEntityId]) -> Any:
        if isinstance(value, TextTranslationReapplier.ToBeTranslated):
            return self.text_translator(value.text, entity_id)
        else:
            return value
    
    def call_impl(self, entity_id: Optional[TextEntityId], obj: Any, field: Hashable, translation_worker: TranslationWorker, text_template: Optional[str], *args, **kwargs):
        new_args = list()
        for arg in args:
            new_args.append(self._translate_needed(arg, entity_id))
        
        new_kwargs = dict()
        for key, value in kwargs.items():
            new_kwargs[key] = self._translate_needed(value, entity_id)
        
        if text_template is None:
            if new_kwargs:
                raise RuntimeError('There are tt items in kwargs, however text_template is None')

            translated_text = ' '.join(new_args)
        else:
            translated_text = text_template.format(*new_args, **new_kwargs)
        
        translation_worker(translated_text)
    
    def args_to_key_value(self, entity_id: Optional[TextEntityId], obj: Any, field: Hashable, translation_worker: Callable, text_template: Optional[str], *args, **kwargs) -> Tuple[Hashable, Any]:
        return ((entity_id, id(obj), field), (translation_worker, text_template, args, kwargs))
    
    def key_value_to_args(self, key: Hashable, value: Any) -> Tuple[Tuple, Dict]:
        entity_id, obj, field = key
        translation_worker, text_template, args, kwargs = value
        new_args = tuple([entity_id, obj, field, translation_worker, text_template] + list(args))
        return new_args, kwargs


def tt(text: TranslatableText) -> 'TextTranslationReapplier.ToBeTranslated':
    return TextTranslationReapplier.ToBeTranslated()(text)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'decode',
    'detect_and_decode',
    'is_utf8_text',
    'is_text_is_7bit_utf8_compatible',
    'is_probably_utf8',
    'is_utf8',
]


from typing import Tuple, Union
import cchardet as chardet
from cengal.modules_management.alternative_import import alt_import
with alt_import('cchardet') as chardet:
    if chardet is None:
        CHARDET_PRESENT: bool = False
    else:
        CHARDET_PRESENT = True

from charset_normalizer import detect as cn_detect
from cengal.text_processing.text_processing import Text, normalize_text
from cengal.text_processing.utf_bom_processing import *


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def detect_and_decode(text: Union[bytes, bytearray], detect_as_utf8_when_possible: bool = True, check_text_for_utf8_compliance: bool = True) -> Tuple[str, str, bytes]:
    if not text:
        return str(), 'utf-8', bytes()

    text = normalize_text(text, bytes)
    possible_utf_bom = determine_text_bom(text)
    text = remove_bom(text, possible_utf_bom)
    possible_encoding = determine_bom_encoding(possible_utf_bom)
    if possible_encoding is not None:
        return decode_text_and_remove_all_wrong_symbols(text, possible_encoding), possible_encoding, possible_utf_bom
    else:
        try_charset_normalizer = False
        try:
            if CHARDET_PRESENT:
                detection = chardet.detect(text)
            else:
                try_charset_normalizer = True
        except LookupError:
            try_charset_normalizer = True
        
        if try_charset_normalizer:
            detection = cn_detect(text)
            
        encoding = detection["encoding"]
        if detect_as_utf8_when_possible:
            if check_text_for_utf8_compliance:
                result_encoding = 'utf-8' if is_utf8_text(encoding, text) else encoding
            else:
                result_encoding = 'utf-8' if (is_utf8(encoding) or is_probably_utf8(encoding)) else encoding
        else:
            result_encoding = encoding
        
        bom_bytes = bytes()
        return text.decode(encoding), result_encoding, bom_bytes


utf8_compatible_encodings = {
    'utf-8',
    'ISO-8859-1',
    'Latin 1',
}
utf8_compatible_encodings_lower = {encoding.lower() for encoding in utf8_compatible_encodings}


utf8_half_compatible_encodings = {
    'US-ASCII',
    'ASCII',
    'ANSI_X3.4-1968',
    'iso-ir-6',
    'ANSI_X3.4-1986',
    'ISO_646.irv:1991',
    'ASCII-7',
    'ASCII-8',
    'ISO646-US',
    'us',
    'IBM367',
    'cp367',
    'csASCII',
}
utf8_half_compatible_encodings_lower = {encoding.lower() for encoding in utf8_half_compatible_encodings}


def is_utf8(encoding: str) -> bool:
    return encoding.lower() in utf8_compatible_encodings_lower


def is_probably_utf8(encoding: str) -> bool:
    return encoding.lower() in utf8_half_compatible_encodings_lower


def is_text_is_7bit_utf8_compatible(text: Union[bytes, bytearray]) -> bool:
    return all(b <= 127 for b in text)


def is_utf8_text(encoding: str, text: Union[bytes, bytearray]) -> bool:
    if is_utf8(encoding):
        return True
    elif is_probably_utf8(encoding):
        return is_text_is_7bit_utf8_compatible(text)
    else:
        return False


def decode(text: Union[bytes, bytearray]) -> str:
    text, encoding, bom_bytes = detect_and_decode(text)
    return text



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
from cengal.parallel_execution.coroutines.coro_tools.loop_administration.admin_tk import start_admin
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_on_keyboard_interrupt import ShutdownOnKeyboardInterrupt
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.instance import InstanceRequest
from cengal.parallel_execution.coroutines.coro_standard_services.log import LogRequest, Log
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.file_system.app_fs_structure.app_dir_path import AppDirPath, AppDirectoryType
from cengal.code_flow_control.args_manager import AK
from cengal.system import current_cengal_module_import_str
from cengal.introspection.inspect import pdi
import os
import tkinter as tk
from tkinter import filedialog


async def on_close(i: Interface):
    await i(ShutdownLoop)


def request_and_set_log_env_dir(i: Interface, admin_tk_app: tk.Tk):
    app_name_for_fs = i(InstanceRequest().wait('app_name_for_fs'))
    app_data_dir_path_type: AppDirectoryType = i(InstanceRequest().wait('app_data_dir_path_type'))
    app_dir_path: AppDirPath = i(InstanceRequest().wait(AppDirPath))
    app_data_dir_path: str = app_dir_path.cached(app_data_dir_path_type, app_name_for_fs, False)
    initialdir = os.path.dirname(app_data_dir_path)
    log_db_env_path = filedialog.askdirectory(parent=admin_tk_app,
                                                initialdir=initialdir,
                                                title="Please select a folder:")
    if not isinstance(log_db_env_path, str):
        i(LogRequest().sync())
        i.log.warning('Path to Log DB Env was not provided! Using own Log DB Env instead.')
        return
    
    log_db_env_path = log_db_env_path.strip()
    log_db_env_path = os.path.normpath(log_db_env_path)
    if not (os.path.exists(log_db_env_path) and os.path.isdir(log_db_env_path)):
        i(LogRequest().sync())
        i.log.warning('Path does not exist! Using own Log DB Env instead.')
        return

    i(LogRequest().set_db_environment_path(log_db_env_path))


async def main(i: Interface):
    await i(ShutdownOnKeyboardInterrupt)
    await i(InstanceRequest().set('app_name_for_fs', current_cengal_module_import_str()))
    await start_admin(i, on_close, AK(current_children_pack_type = 2))
    admin_tk_app: tk.Tk = await i(InstanceRequest().wait('admin_tk_app'))
    await i(RunCoro, request_and_set_log_env_dir, admin_tk_app)


if '__main__' == __name__:
    run_in_loop(main)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['prepare_for_composition']


import ctypes
from ctypes import wintypes
from cengal.system import OS_TYPE


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def prepare_for_composition(hwnd):
    if 'Windows' != OS_TYPE:
        return
    
    dwm = ctypes.windll.dwmapi

    # needs pointertomarginsstruct
    class MARGINS(ctypes.Structure):
        _fields_ = [("cxLeftWidth", ctypes.c_int),
                    ("cxRightWidth", ctypes.c_int),
                    ("cyTopHeight", ctypes.c_int),
                    ("cyBottomHeight", ctypes.c_int)
                    ]
    
    thisWin = hwnd
    
    # if self.is_composition_enabled:
    #     # lt()
    #     margins = MARGINS(-1, 0, 0, 0)
    #     result = dwm.DwmExtendFrameIntoClientArea(thisWin, ctypes.byref(margins))
    # else:
    #     # lt()
    #     margins = MARGINS(0, 0, 0, 0)
    #     result = dwm.DwmExtendFrameIntoClientArea(thisWin, ctypes.byref(margins))
    
    # if 0 == result:
    #     self.is_aero_glass_turned_on = True
    # else:
    #     self.is_aero_glass_turned_on = False

    # --------------------
    
    user32 = ctypes.windll.user32
    
    WCA_ACCENT_POLICY = 19
    ACCENT_ENABLE_BLURBEHIND = 3
    
    class AccentPolicy(ctypes.Structure):
        _fields_ = [
            ('AccentState', wintypes.ULONG),
            ('AccentFlags', wintypes.ULONG),
            ('GradientColor', wintypes.ULONG),
            ('AnimationId', wintypes.ULONG),
        ]
    
    class WINCOMPATTRDATA(ctypes.Structure):
        _fields_ = [
            ('attribute', wintypes.DWORD),
            ('pData', ctypes.POINTER(AccentPolicy)),
            ('dataSize', wintypes.ULONG),
        ]
    
    accent_policy = AccentPolicy(ACCENT_ENABLE_BLURBEHIND, 0, 0, 0)
    wincompattrdata = WINCOMPATTRDATA(WCA_ACCENT_POLICY, ctypes.pointer(accent_policy), ctypes.sizeof(accent_policy))
    
    result = user32.SetWindowCompositionAttribute(thisWin, ctypes.pointer(wincompattrdata))
    
    if 0 != result:
        return None
    else:
        return ctypes.windll.Kernel32.GetLastError()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'EventBus', 'EventBusRequest', 'EventID', 'Handler'
]


from enum import Enum
from typing import Dict, Set, List, Tuple, Union, Type, Optional, Any
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.time_management.repeat_for_a_time import Tracer
from cengal.code_flow_control.smart_values.versions.v_1 import ValueExistence


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


EventID = str
Handler = Callable[[EventID, Any], None]


class EventBusRequest(ServiceRequest):
    def register_handler(self, event: EventID, handler: Handler) -> None:
        return self._save(0, event, handler)

    def remove_handler(self, event: EventID, handler: Handler) -> None:
        return self._save(1, event, handler)

    def send_event(self, event: EventID, data: Any) -> None:
        return self._save(2, event, data)

    def set_priority(self, priority: CoroPriority) -> None:
        return self._save(3, priority)


class EventBus(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(EventBus, self).__init__(loop)

        self._request_workers = {
            0: self._on_register_handler,
            1: self._on_remove_handler,
            2: self._on_send_event,
            3: self._on_set_priority,
        }

        self.handlers: Dict[EventID, Set[Handler]] = dict()
        self.events: Dict[EventID, List[Any]] = dict()
        self.priority: CoroPriority = CoroPriority.low
        self._in_processing: bool = False

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        if EntityStatsMixin.StatsLevel.info == stats_level:
            func_info = full_func_info_to_printable_dict
        else:
            func_info = full_func_info_to_dict

        handlers_set: Set = set()
        for handlers in self.handlers.values():
            handlers_set += handlers
            
        return type(self).__name__, {
            'priority': str(self.priority),
            'events': {
                'num': len(self.events),
                'list': set(self.events)
            },
            'handlers': {
                'num': len(handlers_set),
                'list': [func_info(handler) for handler in handlers_set]
            }
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[EventBusRequest]=None
                                                         ) -> Tuple[bool, Any, None]:
        if request is not None:
            return self.resolve_request(request)
        return True, None, None

    def full_processing_iteration(self):
        if self._in_processing:
            return

        handlers_buff = self.handlers
        self.handlers = type(handlers_buff)()

        events_buff = self.events
        self.events = type(events_buff)()

        interested_events: Set[EventID] = handlers_buff.keys() & events_buff.keys()
        if interested_events:
            async def event_processing_coro(interface: Interface,
                                    interested_events: Set[EventID],
                                    events: Dict[EventID, List[Any]],
                                    handlers: Dict[EventID, Set[Handler]],
                                    priority: CoroPriority
                                    ):
                loop: CoroSchedulerType = interface._loop
                current_coro_interface_buff: Interface = loop.current_coro_interface
                ly = await agly(priority)
                for event in interested_events:
                    event_data_list = events[event]
                    for data in event_data_list:
                        event_handlers = handlers[event]
                        for handler in event_handlers:
                            try:
                                loop.current_coro_interface = None
                                handler(event, data)
                            except:
                                loop.logger.exception('EventBus. Event handler error')
                            finally:
                                loop.current_coro_interface = current_coro_interface_buff
                            
                            await ly()

            coro: CoroWrapperBase = self._loop.put_coro(event_processing_coro,
                                       interested_events, events_buff, handlers_buff,
                                       self.priority)
            # coro.is_background_coro = True  # must not be background coro: it is not an endless coro
            coro.add_on_coro_del_handler(self._on_coro_del_handler)
            self._in_processing = True
        
        self.make_dead()

    def _on_register_handler(self, event: EventID, handler: Handler):
        if event not in self.handlers:
            self.handlers[event] = set()

        self.handlers[event].add(handler)
        self.make_live()
        return True, None, None

    def _on_remove_handler(self, event: EventID, handler: Handler):
        if event in self.handlers:
            self.handlers[event].remove(handler)

        return True, None, None

    def _on_send_event(self, event: EventID, data: Any):
        if event not in self.events:
            self.events[event] = list()

            self.events[event].append(data)
        
        self.make_live()
        return True, None, None

    def _on_set_priority(self, priority: CoroPriority):
        self.priority = priority

        return True, None, None

    def in_work(self):
        result = (not self._in_processing) & bool(self.handlers) & bool(self.events)
        return self.thrifty_in_work(result)

    def _on_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        self._in_processing = False
        if bool(self.handlers) & bool(self.events):
            self.make_live()
        
        return False


EventBusRequest.default_service_type = EventBus


def try_send_event(
        backup_scheduler: Optional[CoroSchedulerType],
        event: EventID, data: Any):
    def event_sender(
            interface: Interface,
            event: EventID, data: Any):
        interface(EventBus, EventBusRequest().send_event(event, data))

    try_put_coro_to(get_interface_and_loop_with_explicit_loop(backup_scheduler), CoroType.auto, event_sender, event, data)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['change_version']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.file_system.file_manager import path_relative_to_current_src
from cengal.text_processing.brackets_processing import BracketPair, Bracket
from cengal.build_tools.ensure_copyright import get_version, set_version
from cengal.os.execute import prepare_py_command
from os import system
from typing import List, Union


default_credits_bracket_pair: BracketPair = BracketPair([Bracket('credits_string =')], [Bracket('credits_string_bracket_pair: BracketPair =')])
default_version_bracket_pair: BracketPair = BracketPair([Bracket('__version__ = "')], [Bracket('"')])


def change_version(credits_string_bracket_pair: BracketPair, version_bracket_pair: BracketPair):
    credits_string_bracket_pair = default_credits_bracket_pair if credits_string_bracket_pair is None else credits_string_bracket_pair
    version_bracket_pair = default_version_bracket_pair if version_bracket_pair is None else version_bracket_pair

    script_file_full_path: str = path_relative_to_current_src('../ensure_copyright/ensure_copyright.py')
    version: str = get_version(script_file_full_path, credits_string_bracket_pair, version_bracket_pair)
    if not version:
        raise ValueError('Version not found')
    
    version_str_parts: List[str] = version.split('.')
    version_parts: List[Union[int, str]] = list()
    for version_part in version_str_parts:
        try:
            version_parts.append(int(version_part))
        except ValueError:
            version_parts.append(version_part)

    print(f'Current version: {version}')
    new_version = input('Enter new version or hit Enter to autoincrease: ').strip()
    if not new_version:
        for reverse_index, version_part in enumerate(reversed(version_parts)):
            if isinstance(version_part, int):
                version_parts[-reverse_index - 1] += 1
                break
        
        new_version = '.'.join(str(version_part) for version_part in version_parts)

    set_version(script_file_full_path, new_version, credits_string_bracket_pair, version_bracket_pair)
    print(f'Version changed to: {new_version}')
    print('Updating version in all files...')
    system(prepare_py_command((script_file_full_path,)))
    print('Done.')



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import codecs
from cengal.text_processing.encoding_detection import detect_and_decode
from cengal.text_processing.text_processing import DEFAULT_ENCODING
from typing import List, Tuple, Optional, Callable, Any



def patch_text_file(
        file_path: str, 
        patch: Any, 
        patch_text: Callable, 
        count: int = 1, 
        encoding: Optional[str] = DEFAULT_ENCODING, 
        normalizer: Optional[Callable] = None,
        fallback_to_utf_8: bool = True,
        replace_errors: bool = True,
        ):
    with open(file_path, 'r+b') as file:
        text, text_encoding, bom_bytes = detect_and_decode(file.read())
        text = patch_text(text, patch, count, encoding, normalizer)
        file.seek(0, 0)
        file.truncate(0)
        try:
            try:
                data = bom_bytes + text.encode(text_encoding)
            except UnicodeEncodeError:
                if fallback_to_utf_8:
                    bom_bytes = codecs.BOM_UTF8
                    text_encoding = 'utf-8'
                    data = bom_bytes + text.encode(text_encoding)
                else:
                    raise
        except UnicodeEncodeError:
            if replace_errors:
                data = bom_bytes + text.encode(text_encoding, errors='replace')
            else:
                raise
        
        file.write(data)


def patch_bin_file(
        file_path: str, 
        patch: Any, 
        patch_text: Callable, 
        count: int = 1, 
        encoding: Optional[str] = DEFAULT_ENCODING, 
        normalizer: Optional[Callable] = None
        ):
    with open(file_path, 'r+b') as file:
        data: bytes = file.read()
        data = patch_text(data, patch, count, encoding, normalizer)
        file.seek(0, 0)
        file.truncate(0)
        file.write(data)
